/**
 *      Copyright (C) 2012 David M. Carr
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

package org.bitbucket.davidm24.mongodb.aggregate;

import com.mongodb.AggregationOutput;
import com.mongodb.BasicDBList;
import com.mongodb.BasicDBObject;
import com.mongodb.DBCollection;
import com.mongodb.DBObject;
import com.mongodb.QueryBuilder;

/**
 * Utility for running aggregate commands.
 * 
 * @author David M. Carr
 * @see <a href="http://docs.mongodb.org/manual/reference/aggregation/">Aggregation Framework Reference</a>
 * @since 0.1.0
 */
public class AggregateBuilder {
    /**
     * A violation of the contract of AggregateBuilder.
     * 
     * @author David M. Carr
     * @since 0.1.1
     */
    @SuppressWarnings("serial")
    public static class AggregateBuilderException extends RuntimeException {
        AggregateBuilderException(String message) {
            super(message);
        }
    }
    
    /**
     * Returns a new AggregateBuilder for the specified collection. All
     * subsequent operators will be treated as part of the processing pipeline.
     * When you're done adding to the pipeline, call {@link #run}.
     * 
     * @param collection the collection within which to aggregate
     * @return the new builder instance
     * @since 0.1.0
     */
    public static AggregateBuilder aggregate(DBCollection collection) {
        return new AggregateBuilder(collection);
    }

    /**
     * Returns a new AggregateBuilder for the specified collection. All
     * subsequent operators will be treated as part of the processing pipeline.
     * When you're done adding to the pipeline, call {@link #get}.  Note that
     * this signature will not support {@link #run}.
     * 
     * @param collectionName the name of the collection within which to
     *            aggregate
     * @return the new builder instance
     * @deprecated use {@link #aggregate(DBCollection)} instead
     * @since 0.1.0
     */
    @Deprecated
    public static AggregateBuilder aggregate(String collectionName) {
        return new AggregateBuilder(collectionName);
    }

    /* For now, optionally provided */
    private DBCollection _collection;

    /* If we remove support for get(), we might not need this any more */
    private DBObject _command;

    private BasicDBList _pipeline;

    /**
     * Instantiates a new AggregateBuilder.
     * 
     * @param collection the collection on which to aggregate
     * @since 0.1.1
     */
    public AggregateBuilder(DBCollection collection) {
        _collection = collection;
        _pipeline = new BasicDBList();
        _command = new BasicDBObject();
        _command.put("aggregate", collection.getName());
        _command.put("pipeline", _pipeline);
    }

    /**
     * Instantiates a new AggregateBuilder.  Note that this signature will not
     * support {@link #run}.
     * 
     * @param collectionName the name of the collection on which to aggregate
     * @deprecated use {@link #AggregateBuilder(DBCollection)} instead
     * @since 0.1.0
     */
    @Deprecated
    public AggregateBuilder(String collectionName) {
        _pipeline = new BasicDBList();
        _command = new BasicDBObject();
        _command.put("aggregate", collectionName);
        _command.put("pipeline", _pipeline);
    }

    private void addOperand(String op, Object value) {
        _pipeline.add(new BasicDBObject(op, value));
    }

    /**
     * Creates a {@link DBObject} to be passed to the driver's command
     * operation.
     * 
     * @return a {@link DBObject} instance
     * @deprecated Use {@link #run} instead
     * @since 0.1.0
     */
    @Deprecated
    public DBObject get() {
        return _command;
    }
    
    /**
     * Adds a {@code $group} operator for the specified builder to the pipeline.
     * 
     * <p>This operator groups documents together for the purpose of calculating
     * aggregate values based on a collection of documents. Practically, group
     * often supports tasks such as average page views for each page in a
     * website on a daily basis.</p>
     * 
     * @param builder a builder specifying how to group
     * @return the AggregateBuilder instance for call chaining
     * @since 0.1.1
     */
    public AggregateBuilder group(GroupBuilder builder) {
        return group(builder.get());
    }

    /**
     * Adds a {@code $group} operator for the specified object to the pipeline.
     * The object should be a document (as a {@link DBObject} or
     * {@link java.util.Map}).
     * 
     * <p>This operator groups documents together for the purpose of calculating
     * aggregate values based on a collection of documents. Practically, group
     * often supports tasks such as average page views for each page in a
     * website on a daily basis.</p>
     * 
     * @param object a document specifying how to group
     * @return the AggregateBuilder instance for call chaining
     * @see GroupBuilder
     * @since 0.1.0
     */
    public AggregateBuilder group(Object object) {
        addOperand(PipelineOperators.GROUP, object);
        return this;
    }

    /**
     * Adds a {@code $group} operator for the specified key/value pair.  The
     * arguments will be treated as a single-key document.  This is intended
     * only as a convenience for very simple group clauses. For more complicated
     * uses, use {@link #group(Object)}.
     * 
     * <p>This operator groups documents together for the purpose of calculating
     * aggregate values based on a collection of documents. Practically, group
     * often supports tasks such as average page views for each page in a
     * website on a daily basis.</p>
     * 
     * @param key the key for the document
     * @param value the value for the document
     * @return the AggregateBuilder instance for call chaining
     * @since 0.1.0
     * @deprecated use {@link #group(GroupBuilder)} instead
     */
    @Deprecated
    public AggregateBuilder group(String key, Object value) {
        return group(new BasicDBObject(key, value));
    }

    /**
     * Adds a {@code $limit} operator for the specified count to the pipeline.
     * 
     * <p>This operator restricts the number of documents that pass through the
     * {@code $limit} in the pipeline.</p>
     * 
     * @param count the maximum number of documents to allow through the
     *          pipeline 
     * @return the AggregateBuilder instance for call chaining
     * @since 0.1.0
     */
    public AggregateBuilder limit(int count) {
        addOperand(PipelineOperators.LIMIT, count);
        return this;
    }
    
    /**
     * Adds a {@code $match} operator for the specified builder to the pipeline.
     * 
     * <p>This operator provides a query-like interface to filter documents out
     * of the aggregation pipeline. The {@code $match} drops documents that do
     * not match the condition from the aggregation pipeline, and it passes
     * documents that match along the pipeline unaltered.  The syntax passed to
     * this operator is identical to the query syntax.</p>
     * 
     * @param builder a builder for the query on which to match
     * @return the AggregateBuilder instance for call chaining
     * @since 0.1.1
     */
    public AggregateBuilder match(QueryBuilder builder) {
        return match(builder.get());
    }

    /**
     * Adds a {@code $match} operator for the specified object to the pipeline.
     * 
     * <p>This operator provides a query-like interface to filter documents out
     * of the aggregation pipeline. The {@code $match} drops documents that do
     * not match the condition from the aggregation pipeline, and it passes
     * documents that match along the pipeline unaltered.  The syntax passed to
     * this operator is identical to the query syntax.</p>
     * 
     * @param object the query on which to match
     * @return the AggregateBuilder instance for call chaining
     * @see com.mongodb.QueryBuilder
     * @since 0.1.0
     */
    public AggregateBuilder match(Object object) {
        addOperand(PipelineOperators.MATCH, object);
        return this;
    }

    /**
     * Adds a {@code $match} operator for the specified key/value pair.  The
     * arguments will be treated as a single-key document.  This is intended
     * only as a convenience for simple single-key, direct equality match
     * clauses. For more complicated uses, use {@link #match(QueryBuilder)}.
     * 
     * <p>This operator provides a query-like interface to filter documents out
     * of the aggregation pipeline. The {@code $match} drops documents that do
     * not match the condition from the aggregation pipeline, and it passes
     * documents that match along the pipeline unaltered.  The syntax passed to
     * this operator is identical to the query syntax.</p>
     * 
     * @param key the field to match
     * @param value the precise value to match to
     * @return the AggregateBuilder instance for call chaining
     * @since 0.1.0
     */
    public AggregateBuilder match(String key, Object value) {
        return match(new BasicDBObject(key, value));
    }

    /**
     * Adds a {@code $project} operator for the specified object to the
     * pipeline.  The object should be a document (as a {@link DBObject} or
     * {@link java.util.Map}).
     * 
     * <p>This operator reshapes a document stream by renaming, adding, or
     * removing fields. Also use {@code $project} to create computed values or
     * sub-objects.</p>
     * 
     * @param object a document specifying how to project
     * @return the AggregateBuilder instance for call chaining
     * @since 0.1.0
     */
    public AggregateBuilder project(Object object) {
        addOperand(PipelineOperators.PROJECT, object);
        return this;
    }

    /**
     * Runs the aggregation pipeline using {@link DBCollection#aggregate} and
     * returns the result.
     * 
     * @return the output of running the aggregation pipeline
     * @throws AggregateBuilderException if you call this method without
     *          initially specifying a {@link DBCollection} instance or without
     *          providing any pipeline operations
     * @since 0.1.1
     */
    public AggregationOutput run() {
        if (_collection == null) {
            throw new AggregateBuilderException("run() is only supported if " +
                    "you initialized the builder with a DBCollection " +
                    "instance.  Otherwise, use get() instead.");
        }
        if (_pipeline.isEmpty()) {
            throw new AggregateBuilderException(
                    "run() requires a non-empty pipeline");
        }
        DBObject firstOp = (DBObject) _pipeline.get(0);
        DBObject[] restOfPipeline = _pipeline.subList(1, _pipeline.size())
                .toArray(new DBObject[_pipeline.size() - 1]);
        return _collection.aggregate(firstOp, restOfPipeline);
    }

    /**
     * Adds a {@code $skip} operator for the specified count to the pipeline.
     * 
     * <p>This operator skips over the specified number of documents that pass
     * through the {@code $skip} in the pipeline before passing all of the
     * remaining input.</p>
     * 
     * @param count the number of documents to skip
     * @return the AggregateBuilder instance for call chaining
     * @since 0.1.0
     */
    public AggregateBuilder skip(int count) {
        addOperand(PipelineOperators.SKIP, count);
        return this;
    }

    /**
     * Adds a {@code $sort} operator for the specified object to the
     * pipeline.  The object should be a document (as a {@link DBObject} or
     * {@link java.util.Map}), with each key being a field to sort by, and each
     * value being {@code 1} or {@code -1} for ascending or descending sort
     * respectively, as in the following example:
     * 
     * <pre>builder.sort(Collections.singletonMap("age", -1))</pre>
     * 
     * <p>This operator sorts all input documents and returns them to the
     * pipeline in sorted order.</p>
     * 
     * @param object a document specifying how to sort
     * @return the AggregateBuilder instance for call chaining
     * @since 0.1.0
     */
    public AggregateBuilder sort(Object object) {
        addOperand(PipelineOperators.SORT, object);
        return this;
    }

    /**
     * Adds a {@code $unwind} operator for the specified source field to the
     * pipeline.
     * 
     * <p>This operator peels off the elements of an array individually, and
     * returns a stream of documents. {@code $unwind} returns one document for
     * every member of the unwound array within every source document.</p>
     * 
     * @param sourceField a reference to the field to unwind
     * @return the AggregateBuilder instance for call chaining
     * @since 0.1.0
     */
    public AggregateBuilder unwind(String sourceField) {
        addOperand(PipelineOperators.UNWIND, sourceField);
        return this;
    }
}
